/* GStreamer
 * Copyright (C) <2005,2006> Wim Taymans <wim@fluendo.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#ifndef __GST_RTSP_WMS_H__
#define __GST_RTSP_WMS_H__

#include <gst/gst.h>

G_BEGIN_DECLS

#define GST_TYPE_RTSP_WMS  		(gst_rtsp_wms_get_type())
#define GST_IS_RTSP_WMS(obj)  		(G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_RTSP_WMS))
#define GST_IS_RTSP_WMS_CLASS(klass) 	(G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_RTSP_WMS))
#define GST_RTSP_WMS(obj)  		(G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_RTSP_WMS, GstRTSPWMS))
#define GST_RTSP_WMS_CLASS(klass)  	(G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_RTSP_WMS, GstRTSPWMSClass))

typedef struct _GstRTSPWMS GstRTSPWMS;
typedef struct _GstRTSPWMSClass GstRTSPWMSClass;

struct _GstRTSPWMS {
  GstElement  element;

  gboolean active;
};

struct _GstRTSPWMSClass {
  GstElementClass parent_class;
};

GType gst_rtsp_wms_get_type(void);

G_END_DECLS

#endif /* __GST_RTSP_WMS_H__ */
